# Local imports
from uplink import hooks


class TestResponseHandler(object):
    def test_handle_response(self, mocker):
        converter = mocker.Mock()
        input_value = "converted"
        converter.return_value = input_value
        rc = hooks.ResponseHandler(converter)
        assert rc.handle_response(None) is input_value


class TestRequestAuditor(object):
    def test_audit_request(self, mocker):
        auditor = mocker.stub()
        ra = hooks.RequestAuditor(auditor)
        ra.audit_request(1, 2, 3)
        auditor.assert_called_with(1, 2, 3)


class TestExceptionHandler(object):
    def test_handle_exception(self, mocker):
        handler = mocker.stub()
        eh = hooks.ExceptionHandler(handler)
        eh.handle_exception(1, 2, 3)
        handler.assert_called_with(1, 2, 3)


class TestTransactionHookChain(object):
    def test_delegate_audit_request(self, transaction_hook_mock):
        chain = hooks.TransactionHookChain(transaction_hook_mock)
        chain.audit_request("method", "url", {})
        transaction_hook_mock.audit_request.assert_called_with(
            "method", "url", {}
        )

    def test_delegate_handle_response(self, transaction_hook_mock):
        chain = hooks.TransactionHookChain(transaction_hook_mock)
        chain.handle_response({})
        transaction_hook_mock.handle_response.assert_called_with({})

    def test_delegate_handle_response_multiple(self, mocker):
        # Include one hook that can't handle responses
        mock_response_handler = mocker.stub()
        mock_request_auditor = mocker.stub()

        chain = hooks.TransactionHookChain(
            hooks.RequestAuditor(mock_request_auditor),
            hooks.ResponseHandler(mock_response_handler),
            hooks.ResponseHandler(mock_response_handler),
        )
        chain.handle_response({})
        mock_response_handler.call_count == 2
        mock_request_auditor.call_count == 1

    def test_delegate_handle_exception(self, transaction_hook_mock):
        chain = hooks.TransactionHookChain(transaction_hook_mock)
        chain.handle_exception(None, None, None)
        transaction_hook_mock.handle_exception.assert_called_with(
            None, None, None
        )



from uplink import helpers


def test_get_api_definitions(request_definition_builder):
    class Service(object):
        builder = request_definition_builder

    assert dict(helpers.get_api_definitions(Service)) == {
        "builder": request_definition_builder
    }


def test_get_api_definitions_from_parent(request_definition_builder):
    class Parent(object):
        builder = request_definition_builder

    class Child(Parent):
        other_builder = request_definition_builder

    assert dict(helpers.get_api_definitions(Child)) == {
        "other_builder": request_definition_builder
    }


class TestRequestBuilder(object):
    def test_return_type(self):
        # Setup
        builder = helpers.RequestBuilder({})

        # Run
        builder.return_type = str

        # Verify
        assert builder.return_type is str

    def test_add_transaction_hook(self, transaction_hook_mock):
        # Setup
        builder = helpers.RequestBuilder({})

        # Run
        builder.add_transaction_hook(transaction_hook_mock)

        # Verify
        assert list(builder.transaction_hooks) == [transaction_hook_mock]



class ExceptionHandler(object):
    def __init__(self, exception_class=Exception):
        self._exc_cls = exception_class
        self._handler = None

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.handle(exc_type, exc_val, exc_tb)

    def set_handler(self, handler):
        self._handler = handler

    def handle(self, exc_type, exc_val, exc_tb):
        if self._handler is not None and isinstance(exc_val, self._exc_cls):
            self._handler(exc_type, exc_val, exc_tb)


class ExceptionHandlerMixin(object):
    __exception_handler = None

    @property
    def _exception_handler(self):
        if self.__exception_handler is None:
            self.__exception_handler = ExceptionHandler()
        return self.__exception_handler

    def add_exception_handler(self, exception_handler):
        self._exception_handler.set_handler(exception_handler)



# Standard library imports
import typing

# Third party imports
import marshmallow
import pytest

# Local imports
from uplink import converters
from uplink.converters import register, standard


class TestCast(object):
    def test_converter_without_caster(self, mocker):
        converter_mock = mocker.stub()
        converter_mock.return_value = 2
        cast = standard.Cast(None, converter_mock)
        return_value = cast.convert(1)
        converter_mock.assert_called_with(1)
        assert return_value == 2

    def test_convert_with_caster(self, mocker):
        caster = mocker.Mock(return_value=2)
        converter_mock = mocker.Mock(return_value=3)
        cast = standard.Cast(caster, converter_mock)
        return_value = cast.convert(1)
        caster.assert_called_with(1)
        converter_mock.assert_called_with(2)
        assert return_value == 3


class TestRequestBodyConverter(object):
    def test_convert_str(self):
        converter_ = standard.RequestBodyConverter()
        assert converter_.convert("example") == "example"

    def test_convert_obj(self):
        converter_ = standard.RequestBodyConverter()
        example = {"hello": "2"}
        assert converter_.convert(example) == example


class TestStringConverter(object):
    def test_convert(self):
        converter_ = standard.StringConverter()
        assert converter_.convert(2) == "2"


class TestStandardConverter(object):
    def test_create_response_body_converter(self, converter_mock):
        # Setup
        factory = standard.StandardConverter()

        # Run & Verify: Pass-through converters
        converter = factory.create_response_body_converter(converter_mock)
        assert converter is converter_mock

        # Run & Verify: Otherwise, return None
        assert None is factory.create_response_body_converter("converter")


class TestConverterFactoryRegistry(object):
    backend = converters.ConverterFactoryRegistry._converter_factory_registry

    def test_init_args_are_passed_to_factory(
        self, converter_factory_mock, converter_mock
    ):
        args = ("arg1", "arg2")
        kwargs = {"arg3": "arg3"}
        converter_factory_mock.create_string_converter.return_value = (
            converter_mock
        )
        registry = converters.ConverterFactoryRegistry(
            (converter_factory_mock,), *args, **kwargs
        )
        return_value = registry[converters.keys.CONVERT_TO_STRING]()
        converter_factory_mock.create_string_converter.assert_called_with(
            *args, **kwargs
        )
        assert return_value is converter_mock

        # Test with type that can't be handled by registry
        converter_factory_mock.create_string_converter.return_value = None
        return_value = registry[converters.keys.CONVERT_TO_STRING]()
        assert return_value is None

    def test_hooks(self, converter_factory_mock, converter_mock):
        converter_factory_mock.create_string_converter.return_value = (
            converter_mock
        )
        registry = converters.ConverterFactoryRegistry(
            (converter_factory_mock,)
        )
        registry[converters.keys.CONVERT_TO_STRING]()
        assert converter_mock.set_chain.called

    def test_len(self):
        registry = converters.ConverterFactoryRegistry(())
        assert len(registry) == len(self.backend)

    def test_iter(self):
        registry = converters.ConverterFactoryRegistry(())
        assert list(iter(registry)) == list(iter(self.backend))


def test_create_request_body_converter(converter_factory_mock):
    method = converters.create_request_body_converter(converter_factory_mock)
    assert method is converter_factory_mock.create_request_body_converter


def test_create_response_body_converter(converter_factory_mock):
    method = converters.create_response_body_converter(converter_factory_mock)
    assert method is converter_factory_mock.create_response_body_converter


def test_create_string_converter(converter_factory_mock):
    method = converters.create_string_converter(converter_factory_mock)
    assert method is converter_factory_mock.create_string_converter


@pytest.fixture(params=["class", "instance"])
def schema_mock_and_argument(request, mocker):
    class Schema(marshmallow.Schema):
        def __new__(cls, *args, **kwargs):
            return schema

    schema = mocker.Mock(spec=marshmallow.Schema)
    if request.param == "class":
        return schema, Schema
    else:
        return schema, schema


class TestMarshmallowConverter(object):
    def test_init_without_marshmallow(self):
        old_marshmallow = converters.MarshmallowConverter.marshmallow
        converters.MarshmallowConverter.marshmallow = None
        with pytest.raises(ImportError):
            converters.MarshmallowConverter()
        converters.MarshmallowConverter.marshmallow = old_marshmallow

    def test_create_request_body_converter(
        self, mocker, schema_mock_and_argument
    ):
        # Setup
        schema_mock, argument = schema_mock_and_argument
        expected_result = "data"
        dump_result = mocker.Mock()
        dump_result.data = expected_result
        schema_mock.dump.return_value = dump_result
        converter = converters.MarshmallowConverter()
        request_body = {"id": 0}

        # Run
        c = converter.create_request_body_converter(argument)
        result = c.convert(request_body)

        # Verify
        schema_mock.dump.assert_called_with(request_body)
        assert expected_result == result

    def test_create_request_body_converter_without_schema(self):
        # Setup
        converter = converters.MarshmallowConverter()

        # Run
        c = converter.create_request_body_converter("not a schema")

        # Verify
        assert c is None

    def test_create_response_body_converter(
        self, mocker, schema_mock_and_argument
    ):
        # Setup
        schema_mock, argument = schema_mock_and_argument
        expected_result = "data"
        load_result = mocker.Mock()
        load_result.data = expected_result
        schema_mock.load.return_value = load_result
        converter = converters.MarshmallowConverter()
        response = mocker.Mock(spec=["json"])
        c = converter.create_response_body_converter(argument)

        # Run & Verify: with response
        result = c.convert(response)
        response.json.assert_called_with()
        schema_mock.load.assert_called_with(response.json())
        assert expected_result == result

        # Run & Verify: with json
        data = {"hello": "world"}
        result = c.convert(data)
        schema_mock.load.assert_called_with(data)
        assert expected_result == result

        # Run & Verify: raise validation errors for user to handle.
        schema_mock.load.side_effect = marshmallow.exceptions.MarshmallowError

        with pytest.raises(marshmallow.exceptions.MarshmallowError):
            c.convert(data)

    def test_create_response_body_converter_with_unsupported_response(
        self, schema_mock_and_argument
    ):
        # Setup
        schema_mock, argument = schema_mock_and_argument
        converter = converters.MarshmallowConverter()

        # Run
        c = converter.create_response_body_converter(argument)
        result = c.convert("unsupported response")

        # Verify
        return result is None

    def test_create_response_body_converter_without_schema(self):
        # Setup
        converter = converters.MarshmallowConverter()

        # Run
        c = converter.create_response_body_converter("not a schema")

        # Verify
        assert c is None

    def test_create_string_converter(self, schema_mock_and_argument):
        # Setup
        _, argument = schema_mock_and_argument
        converter = converters.MarshmallowConverter()

        # Run
        c = converter.create_string_converter(argument, None)

        # Verify
        assert c is None

    def test_register(self, mocker):
        # Setup
        converter = converters.MarshmallowConverter
        old_marshmallow = converter.marshmallow

        # Run & Verify: Register when marshmallow is installed
        converter.marshmallow = True
        register_ = []
        converter.register_if_necessary(register_.append)
        assert register_ == [converter]

        # Run & Verify: Skip when marshmallow is not installed
        converter.marshmallow = None
        register_ = []
        converter.register_if_necessary(register_.append)
        assert register_ == []

        # Rewire
        converters.MarshmallowConverter.marshmallow = old_marshmallow


class TestMap(object):
    def test_convert(self):
        # Setup
        registry = converters.ConverterFactoryRegistry(
            (converters.StandardConverter(),)
        )
        key = converters.keys.Map(converters.keys.CONVERT_TO_STRING)

        # Run
        converter = registry[key](None)
        value = converter({"hello": 1})

        # Verify
        assert value == {"hello": "1"}

    def test_eq(self):
        assert converters.keys.Map(0) == converters.keys.Map(0)
        assert not (converters.keys.Map(1) == converters.keys.Map(0))
        assert not (converters.keys.Map(1) == 1)


class TestSequence(object):
    def test_convert_with_sequence(self):
        # Setup
        registry = converters.ConverterFactoryRegistry(
            (converters.StandardConverter(),)
        )
        key = converters.keys.Sequence(converters.keys.CONVERT_TO_STRING)

        # Run
        converter = registry[key](None)
        value = converter([1, 2, 3])

        # Verify
        assert value == ["1", "2", "3"]

    def test_convert_not_sequence(self):
        # Setup
        registry = converters.ConverterFactoryRegistry(
            (converters.StandardConverter(),)
        )
        key = converters.keys.Sequence(converters.keys.CONVERT_TO_STRING)

        # Run
        converter = registry[key](None)
        value = converter("1")

        # Verify
        assert value == "1"

    def test_eq(self):
        assert converters.keys.Sequence(0) == converters.keys.Sequence(0)
        assert not (converters.keys.Sequence(1) == converters.keys.Sequence(0))
        assert not (converters.keys.Sequence(1) == 1)


class TestRegistry(object):
    @pytest.mark.parametrize(
        "converter",
        # Try with both class and instance
        (converters.StandardConverter, converters.StandardConverter()),
    )
    def test_register_converter_factory_pass(self, converter):
        # Setup
        registry = register.Register()

        # Verify
        return_value = registry.register_converter_factory(converter)
        defaults = registry.get_converter_factories()
        assert return_value is converter
        assert len(defaults) == 1
        assert isinstance(defaults[0], converters.StandardConverter)

    def test_register_converter_factory_fail(self):
        # Setup
        registry = register.Register()

        # Verify failure when registered factory is not proper type.
        with pytest.raises(TypeError):
            registry.register_converter_factory(object())


class TestTypingConverter(object):
    singleton = converters.TypingConverter()
    inject_methods = pytest.mark.parametrize(
        "method, use_typing",
        [
            (singleton.create_request_body_converter, True),
            (singleton.create_request_body_converter, False),
            (singleton.create_response_body_converter, True),
            (singleton.create_response_body_converter, False),
        ],
    )

    @inject_methods
    def test_methods(self, method, use_typing):
        List, Dict = converters.typing_._get_types(use_typing)

        # Verify with sequence
        converter = method(List[str])
        assert isinstance(converter, converters.typing_.ListConverter)

        # Verify with mapping
        converter = method(Dict[str, str])
        assert isinstance(converter, converters.typing_.DictConverter)

        # Verify with unsupported value
        converter = method(None)
        assert converter is None

        # Verify with unsupported type
        if use_typing:
            converter = method(typing.Set[str])
            assert converter is None

    def test_list_converter(self):
        # Setup
        converter = converters.typing_.ListConverter(str)
        converter.set_chain(lambda x: x)

        # Verify
        output = converter([1, 2, 3])
        assert output == ["1", "2", "3"]

        # Verify with non-list: use element converter
        output = converter(1)
        assert output == ["1"]

    def test_dict_converter(self):
        # Setup
        converter = converters.typing_.DictConverter(int, str)
        converter.set_chain(lambda x: x)

        # Verify
        output = converter({"1": 1, "2": 2})
        assert output == {1: "1", 2: "2"}

        # Verify with non-map: use value converter
        output = converter(1)
        assert output == "1"



# Third party imports
import pytest

# Local imports
from uplink import auth


class TestGetAuth(object):
    def test_tuple(self):
        authentication = auth.get_auth(("username", "password"))
        assert isinstance(authentication, auth.BasicAuth)

    def test_callable(self):
        def func():
            None

        output = auth.get_auth(func)
        assert output is func

    def test_unsupported(self):
        with pytest.raises(ValueError):
            auth.get_auth(object())


def test_basic_auth(request_builder):
    # Setup
    basic_auth = auth.BasicAuth("username", "password")

    # Verify
    basic_auth(request_builder)
    auth_str = basic_auth._auth_str
    assert request_builder.info["headers"]["Authorization"] == auth_str


def test_proxy_auth(request_builder):
    # Setup
    proxy_auth = auth.ProxyAuth("username", "password")

    # Verify
    proxy_auth(request_builder)
    auth_str = proxy_auth._auth_str
    assert request_builder.info["headers"]["Proxy-Authorization"] == auth_str


def test_bearer_token(request_builder):
    # Setup
    bearer_token = auth.BearerToken("bearer-token")

    # Verify
    bearer_token(request_builder)
    auth_str = bearer_token._auth_str
    assert request_builder.info["headers"]["Authorization"] == auth_str



# Standard library imports
import sys

# Local imports
from uplink import decorators
from uplink.converters import keys, interfaces

__all__ = ["json", "from_json", "model"]


class _ReturnsBase(decorators.MethodAnnotation):
    def _get_return_type(self, return_type):  # pragma: no cover
        return return_type

    def _make_strategy(self, converter):  # pragma: no cover
        pass

    def modify_request(self, request_builder):
        return_type = self._get_return_type(request_builder.return_type)
        if isinstance(return_type, _StrategyWrapper):
            converter = return_type.unwrap()
        else:
            converter = request_builder.get_converter(
                keys.CONVERT_FROM_RESPONSE_BODY, return_type
            )
        if converter is not None:
            # Found a converter that can handle the return type.
            request_builder.return_type = _StrategyWrapper(
                converter, self._make_strategy(converter)
            )


class _StrategyWrapper(object):
    def __init__(self, converter, strategy):
        self._converter = converter
        self._strategy = strategy

    def __call__(self, *args, **kwargs):
        return self._strategy(*args, **kwargs)

    def unwrap(self):  # pragma: no cover
        return self._converter


class JsonStrategy(object):
    # TODO: Consider moving this under json decorator
    # TODO: Support JSON Pointer (https://tools.ietf.org/html/rfc6901)

    def __init__(self, converter, member=()):
        self._converter = converter

        if not isinstance(member, (list, tuple)):
            member = (member,)
        self._member = member

    def __call__(self, response):
        content = response.json()
        for name in self._member:
            content = content[name]
        content = self._converter(content)
        return content

    def unwrap(self):
        return self._converter


# noinspection PyPep8Naming
class json(_ReturnsBase):
    """
    Specifies that the decorated consumer method should return a JSON
    object.

    .. code-block:: python

        # This method will return a JSON object (e.g., a dict or list)
        @returns.json
        @get("/users/{username}")
        def get_user(self, username):
            \"""Get a specific user.\"""

    Returning a Specific JSON Field:

        The :py:attr:`member` argument accepts a string or tuple that
        specifies the path of an internal field in the JSON document.

        For instance, consider an API that returns JSON responses that,
        at the root of the document, contains both the server-retrieved
        data and a list of relevant API errors:

        .. code-block:: json
            :emphasize-lines: 2

            {
                "data": { "user": "prkumar", "id": 140232 },
                "errors": []
            }

        If returning the list of errors is unnecessary, we can use the
        :py:attr:`member` argument to strictly return the inner field
        :py:attr:`data`:

        .. code-block:: python

            @returns.json(member="data")
            @get("/users/{username}")
            def get_user(self, username):
                \"""Get a specific user.\"""

    .. versionadded:: v0.5.0
    """

    _can_be_static = True

    class _DummyConverter(interfaces.Converter):
        def convert(self, response):
            return response

    __dummy_converter = _DummyConverter()

    def __init__(self, model=None, member=()):
        # TODO: Consider renaming `model` and `member` to `schema` and
        # `key`, respectively.
        self._model = model
        self._member = member

    def _get_return_type(self, return_type):
        # If the model and return type are None, the strategy should
        # directly return the JSON body of the HTTP response, instead of
        # trying to deserialize it into a model. In this case, by
        # defaulting the return type to the dummy converter, which
        # implements this pass-through behavior, we ensure that
        # _make_strategy is called.
        default = self.__dummy_converter if self._model is None else self._model

        return default if return_type is None else return_type

    def _make_strategy(self, converter):
        return JsonStrategy(converter, self._member)


from_json = json
"""
    Specifies that the decorated consumer method should produce
    instances of a :py:obj:`model` class using a registered
    deserialization strategy (see :py:meth:`uplink.loads.from_json`)

    This decorator accepts the same arguments as
    :py:class:`uplink.returns.json`.

    Often, a JSON response body represents a model in your application.
    If an existing Python object encapsulates this model, use the
    :py:attr:`model` argument to specify it as the return type:

    .. code-block:: python

        @returns.from_json(model=User)
        @get("/users/{username}")
        def get_user(self, username):
            \"""Get a specific user.\"""

    For Python 3 users, you can alternatively provide a return value
    annotation. Hence, the previous code is equivalent to the following
    in Python 3:

    .. code-block:: python

        @returns.from_json
        @get("/users/{username}")
        def get_user(self, username) -> User:
            \"""Get a specific user.\"""

    Both usages typically require also registering a converter that
    knows how to deserialize the JSON into your data model object (see
    :py:meth:`uplink.loads.from_json`). This step is unnecessary if
    these objects are defined using a library for whom Uplink has
    built-in support, such as :py:mod:`marshmallow`.

    .. versionadded:: v0.6.0

"""


# noinspection PyPep8Naming
class model(_ReturnsBase):
    """
    Specifies that the function returns a specific class.

    In Python 3, to provide a consumer method's return type, you can
    set it as the method's return annotation:

    .. code-block:: python

        @get("/users/{username}")
        def get_user(self, username) -> UserSchema:
            \"""Get a specific user.\"""

    For Python 2.7 compatibility, you can use this decorator instead:

    .. code-block:: python

        @returns.model(UserSchema)
        @get("/users/{username}")
        def get_user(self, username):
            \"""Get a specific user.\"""

    To have Uplink convert response bodies into the desired type, you
    will need to define an appropriate converter (e.g., using
    :py:class:`uplink.loads`).

    .. versionadded:: v0.5.1
    """

    def __init__(self, type):
        self._type = type

    def _get_return_type(self, return_type):
        return self._type if return_type is None else return_type

    def _make_strategy(self, converter):
        return converter


class _ModuleProxy(object):
    __module = sys.modules[__name__]

    model = model
    json = json
    from_json = from_json
    __all__ = __module.__all__

    def __getattr__(self, item):
        return getattr(self.__module, item)

    def __call__(self, *args, **kwargs):
        return model(*args, **kwargs)


sys.modules[__name__] = _ModuleProxy()



# Standard library imports
import contextlib

# Third party imports
import pytest

# Local imports
from uplink.clients import (
    AiohttpClient,
    interfaces,
    requests_,
    twisted_,
    register,
)

try:
    from uplink.clients import aiohttp_
except (ImportError, SyntaxError):
    aiohttp_ = None


requires_python34 = pytest.mark.skipif(
    not aiohttp_, reason="Requires Python 3.4 or above"
)


@contextlib.contextmanager
def _patch(obj, attr, value):
    if obj is not None:
        old_value = getattr(obj, attr)
        setattr(obj, attr, value)
    yield
    if obj is not None:
        setattr(obj, attr, old_value)


def test_get_default_client_with_non_callable():
    # Setup
    old_default = register.get_default_client()
    register.set_default_client("client")
    default_client = register.get_default_client()
    register.set_default_client(old_default)

    # Verify: an object that is not callable should be returned as set.
    assert default_client == "client"


def test_get_client_with_http_client_adapter_subclass():
    class HttpClientAdapterMock(interfaces.HttpClientAdapter):
        def create_request(self):
            pass

    client = register.get_client(HttpClientAdapterMock)
    assert isinstance(client, HttpClientAdapterMock)


def test_get_client_with_unrecognized_key():
    assert register.get_client("no client for this key") is None


class TestRequests(object):
    def test_get_client(self, mocker):
        import requests

        session_mock = mocker.Mock(spec=requests.Session)
        client = register.get_client(session_mock)
        assert isinstance(client, requests_.RequestsClient)

    def test_init_with_kwargs(self):
        session = requests_.RequestsClient._create_session(
            cred=("username", "password")
        )
        assert session.cred == ("username", "password")

    def test_create_request(self):
        client = requests_.RequestsClient()
        request = client.create_request()
        assert isinstance(request, requests_.Request)

    def test_request_send(self, mocker):
        # Setup
        import requests

        session_mock = mocker.Mock(spec=requests.Session)
        session_mock.request.return_value = "response"
        callback = mocker.stub()
        client = requests_.Request(session_mock)

        # Run
        client.send("method", "url", {})

        # Verify
        session_mock.request.assert_called_with(method="method", url="url")

        # Run with callback
        client.add_callback(callback)
        client.send("method", "url", {})
        session_mock.request.assert_called_with(method="method", url="url")
        callback.assert_called_with(session_mock.request.return_value)


class TestTwisted(object):
    def test_init_without_client(self):
        twisted = twisted_.TwistedClient()
        assert isinstance(twisted._requests, requests_.RequestsClient)

    def test_create_requests(self, http_client_mock):
        twisted = twisted_.TwistedClient(http_client_mock)
        request = twisted.create_request()
        assert request._proxy is http_client_mock.create_request()
        assert isinstance(request, twisted_.Request)

    def test_create_requests_no_twisted(self, http_client_mock):
        with _patch(twisted_, "threads", None):
            with pytest.raises(NotImplementedError):
                twisted_.TwistedClient(http_client_mock)

    def test_request_send(self, mocker, request_mock):
        deferToThread = mocker.patch.object(twisted_.threads, "deferToThread")
        request = twisted_.Request(request_mock)
        request.send(1, 2, 3)
        deferToThread.assert_called_with(request_mock.send, 1, 2, 3)

    def test_request_send_callback_and_exception(self, mocker, request_mock):
        # Setup
        callback = mocker.stub()
        exception_handler = mocker.stub()
        deferred = mocker.Mock()
        deferToThread = mocker.patch.object(twisted_.threads, "deferToThread")
        deferToThread.return_value = deferred
        request = twisted_.Request(request_mock)
        request.add_callback(callback)
        request.add_exception_handler(exception_handler)

        # Run
        request.send(1, 2, 3)

        # Verify
        deferred.addCallback.assert_called_with(callback)
        deferred.addErrback(request.handle_failure)
        deferToThread.assert_called_with(request_mock.send, 1, 2, 3)

    def test_handle_failure(self, mocker, request_mock):
        # Setup
        failure = mocker.Mock(stub=twisted_.threads.failure.Failure)
        failure.type, failure.value = Exception, Exception()
        exception_handler = mocker.stub()
        request = twisted_.Request(request_mock)
        request.add_exception_handler(exception_handler)

        # Run
        request.handle_failure(failure)

        # Verify
        exception_handler.assert_called_with(
            failure.type, failure.value, failure.getTracebackObject()
        )


@pytest.fixture
def aiohttp_session_mock(mocker):
    import aiohttp

    return mocker.Mock(spec=aiohttp.ClientSession)


class TestAiohttp(object):
    @requires_python34
    def test_init_when_aiohttp_is_not_installed(self):
        with _patch(aiohttp_, "aiohttp", None):
            with pytest.raises(NotImplementedError):
                AiohttpClient()

    @requires_python34
    def test_init_with_session_None(self, mocker):
        mocker.spy(AiohttpClient, "_create_session")
        AiohttpClient(kwarg1="value")
        AiohttpClient._create_session.assert_called_with(kwarg1="value")

    @requires_python34
    def test_get_client(self, aiohttp_session_mock):
        client = register.get_client(aiohttp_session_mock)
        assert isinstance(client, aiohttp_.AiohttpClient)

    @requires_python34
    def test_create_request(self, aiohttp_session_mock):
        aiohttp = aiohttp_.AiohttpClient(aiohttp_session_mock)
        assert isinstance(aiohttp.create_request(), aiohttp_.Request)

    @requires_python34
    def test_request_send(self, aiohttp_session_mock):
        # Setup
        import asyncio

        @asyncio.coroutine
        def request(*args, **kwargs):
            return 0

        aiohttp_session_mock.request = request
        client = aiohttp_.AiohttpClient(aiohttp_session_mock)
        request = aiohttp_.Request(client)

        # Run
        response = request.send(1, 2, {})
        loop = asyncio.get_event_loop()
        value = loop.run_until_complete(asyncio.ensure_future(response))

        # Verify
        assert value == 0

    @requires_python34
    def test_callback(self, aiohttp_session_mock):
        # Setup
        import asyncio

        @asyncio.coroutine
        def request(*args, **kwargs):
            return 2

        aiohttp_session_mock.request = request
        client = aiohttp_.AiohttpClient(aiohttp_session_mock)
        client._sync_callback_adapter = asyncio.coroutine
        request = aiohttp_.Request(client)

        # Run
        request.add_callback(lambda x: 2)
        response = request.send(1, 2, {})
        loop = asyncio.get_event_loop()
        value = loop.run_until_complete(asyncio.ensure_future(response))

        # Verify
        assert value == 2

    @requires_python34
    def test_wrap_callback(self, mocker):
        import asyncio

        # Setup
        c = AiohttpClient()
        mocker.spy(c, "_sync_callback_adapter")

        # Run: with callback that is not a coroutine
        def callback(*_):
            pass

        c.wrap_callback(callback)

        # Verify: Should wrap it
        c._sync_callback_adapter.assert_called_with(callback)

        # Run: with coroutine callback
        coroutine_callback = asyncio.coroutine(callback)
        assert c.wrap_callback(coroutine_callback) is coroutine_callback

    @requires_python34
    def test_threaded_callback(self, mocker):
        import asyncio

        def callback(response):
            return response

        # Mock response.
        response = mocker.Mock()
        response.text = asyncio.coroutine(mocker.stub())

        # Run
        new_callback = aiohttp_.threaded_callback(callback)
        return_value = new_callback(response)
        loop = asyncio.get_event_loop()
        value = loop.run_until_complete(asyncio.ensure_future(return_value))

        # Verify
        response.text.assert_called_with()
        assert value == response

        # Run: Verify with callback that returns new value
        def callback(*_):
            return 1

        new_callback = aiohttp_.threaded_callback(callback)
        return_value = new_callback(response)
        loop = asyncio.get_event_loop()
        value = loop.run_until_complete(asyncio.ensure_future(return_value))
        assert value == 1

    @requires_python34
    def test_threaded_coroutine(self):
        # Setup
        import asyncio

        @asyncio.coroutine
        def coroutine():
            return 1

        threaded_coroutine = aiohttp_.ThreadedCoroutine(coroutine)

        # Run -- should block
        response = threaded_coroutine()

        # Verify
        assert response == 1

    @requires_python34
    def test_threaded_response(self, mocker):
        # Setup
        import asyncio

        @asyncio.coroutine
        def coroutine():
            return 1

        def not_a_coroutine():
            return 2

        response = mocker.Mock()
        response.coroutine = coroutine
        response.not_coroutine = not_a_coroutine
        threaded_response = aiohttp_.ThreadedResponse(response)

        # Run
        threaded_coroutine = threaded_response.coroutine
        return_value = threaded_coroutine()

        # Verify
        assert isinstance(threaded_coroutine, aiohttp_.ThreadedCoroutine)
        assert return_value == 1
        assert threaded_response.not_coroutine is not_a_coroutine

    @requires_python34
    def test_create(self, mocker):
        # Setup
        import asyncio

        session_cls_mock = mocker.patch("aiohttp.ClientSession")
        positionals = [1]
        keywords = {"keyword": 2}

        # Run: Create client
        client = aiohttp_.AiohttpClient.create(*positionals, **keywords)

        # Verify: session hasn't been created yet.
        assert not session_cls_mock.called

        # Run: Get session
        loop = asyncio.get_event_loop()
        loop.run_until_complete(asyncio.ensure_future(client.session()))

        # Verify: session created with args
        session_cls_mock.assert_called_with(*positionals, **keywords)



# Third party imports
import pytest
import requests

# Local imports.
import uplink
from uplink import utils, clients

# Constants
BASE_URL = "https://api.github.com/"


def _get_url(url):
    return utils.urlparse.urljoin(BASE_URL, url)


@uplink.timeout(10)
@uplink.headers({"Accept": "application/vnd.github.v3.full+json"})
class GitHubService(uplink.Consumer):
    @uplink.timeout(15)
    @uplink.get("/users/{user}/repos")
    def list_repos(self, user):
        pass

    @uplink.returns.json
    @uplink.get("/users/{user}/repos/{repo}")
    def get_repo(self, user, repo):
        pass


@pytest.fixture
def github_service_and_client(http_client_mock):
    return (
        GitHubService(base_url=BASE_URL, client=http_client_mock),
        http_client_mock,
    )


def test_list_repo(github_service_and_client):
    service, http_client = github_service_and_client
    service.list_repos("prkumar")
    http_client.create_request().send.assert_called_with(
        "GET",
        _get_url("/users/prkumar/repos"),
        {
            "headers": {"Accept": "application/vnd.github.v3.full+json"},
            "timeout": 15,
        },
    )


class RequestMock(clients.interfaces.Request):
    def __init__(self, response):
        self._response = response
        self._callback = None

    def send(self, method, url, extras):
        if self._callback is not None:
            return self._callback(self._response)
        else:
            return self._response

    def add_callback(self, callback):
        self._callback = callback

    def add_exception_handler(self, exception_handler):
        pass


def test_get_repo(github_service_and_client, mocker):
    """
    This integration test ensures that the returns.json returns the
    Json body when a model is not provided.
    """
    # Setup: return a mock response
    service, http_client = github_service_and_client
    response = mocker.Mock(spec=requests.Response)
    expected_body = response.json.return_value = {"key": "value"}
    http_client.create_request.return_value = RequestMock(response)

    # Run
    result = service.get_repo("prkumar", "uplink")

    # Verify
    assert expected_body == result



# Standard library imports
import atexit

# Third party imports
import requests

# Local imports
from uplink.clients import helpers, interfaces, register


class RequestsClient(interfaces.HttpClientAdapter):
    """
    A :py:mod:`requests` client that returns
    :py:class:`requests.Response` responses.

    Args:
        session (:py:class:`requests.Session`, optional): The session
            that should handle sending requests. If this argument is
            omitted or set to :py:obj:`None`, a new session will be
            created.
    """

    def __init__(self, session=None, **kwargs):
        if session is None:
            session = self._create_session(**kwargs)
        self.__session = session

    def create_request(self):
        return Request(self.__session)

    @staticmethod
    @register.handler
    def with_session(session, *args, **kwargs):
        if isinstance(session, requests.Session):
            return RequestsClient(session, *args, **kwargs)

    @staticmethod
    def _create_session(**kwargs):
        session = requests.Session()
        atexit.register(session.close)
        for key in kwargs:
            setattr(session, key, kwargs[key])
        return session


class Request(helpers.ExceptionHandlerMixin, interfaces.Request):
    def __init__(self, session):
        self._session = session
        self._callback = None

    def send(self, method, url, extras):
        with self._exception_handler:
            response = self._session.request(method=method, url=url, **extras)
        if self._callback is not None:
            response = self._callback(response)
        return response

    def add_callback(self, callback):
        self._callback = callback



# Local imports
from uplink import returns


def test_returns(request_builder):
    request_builder.get_converter.return_value = str
    custom = returns(str)
    custom.modify_request(request_builder)
    assert request_builder.return_type.unwrap() is str


def test_returns_json(request_builder):
    request_builder.get_converter.return_value = str
    returns_json = returns.json(str, ())
    returns_json.modify_request(request_builder)
    assert isinstance(request_builder.return_type, returns._StrategyWrapper)

    # Verify: Idempotent
    returns_json.modify_request(request_builder)
    assert isinstance(request_builder.return_type, returns._StrategyWrapper)
    assert request_builder.return_type.unwrap() is str

    # Verify: Doesn't apply to unsupported types
    request_builder.get_converter.return_value = None
    request_builder.return_type = None
    returns_json = returns.json(str, ())
    returns_json.modify_request(request_builder)
    assert request_builder.return_type is None


def test_StrategyWrapper():
    wrapper = returns._StrategyWrapper(None, lambda x: x)
    assert wrapper("hello") == "hello"


def test_returns_JsonStrategy(mocker):
    response = mocker.Mock(spec=["json"])
    response.json.return_value = {"hello": "world"}
    converter = returns.JsonStrategy(lambda x: x, "hello")
    assert converter(response) == "world"

    converter = returns.JsonStrategy(lambda y: y + "!", "hello")
    assert converter(response) == "world!"

    assert returns.JsonStrategy(1).unwrap() == 1



# Standard library imports
import json

# Local imports
from uplink.converters import interfaces, register_default_converter_factory


class Cast(interfaces.Converter):
    def __init__(self, caster, converter):
        self._cast = caster
        self._converter = converter

    def set_chain(self, chain):
        self._converter.set_chain(chain)

    def convert(self, value):
        if callable(self._cast):
            value = self._cast(value)
        return self._converter(value)


class RequestBodyConverter(interfaces.Converter):
    @staticmethod
    def _default_json_dumper(obj):
        return obj.__dict__  # pragma: no cover

    def convert(self, value):
        if isinstance(value, str):
            return value
        dumped = json.dumps(value, default=self._default_json_dumper)
        return json.loads(dumped)


class StringConverter(interfaces.Converter):
    def convert(self, value):
        return str(value)


@register_default_converter_factory
class StandardConverter(interfaces.Factory):
    """
    The default converter, this class seeks to provide sane alternatives
    for (de)serialization when all else fails -- e.g., no other
    converters could handle a particular type.
    """

    def create_response_body_converter(self, type_, *args, **kwargs):
        if isinstance(type_, interfaces.Converter):
            return type_

    def create_request_body_converter(self, type_, *args, **kwargs):
        return Cast(type_, RequestBodyConverter())  # pragma: no cover

    def create_string_converter(self, type_, *args, **kwargs):
        return Cast(type_, StringConverter())  # pragma: no cover
